/**
 * \file jdir.hpp
 *
 * Qpid asynchronous store plugin library
 *
 * File containing code for class mrg::journal::jdir (journal data
 * directory), used for controlling and manipulating journal data
 * directories and files. See class documentation for details.
 *
 * \author Kim van der Riet
 *
 * Copyright (c) 2007, 2008 Red Hat, Inc.
 *
 * This file is part of the Qpid async store library msgstore.so.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
 * USA
 *
 * The GNU Lesser General Public License is available in the file COPYING.
 */

#ifndef mrg_journal_jdir_hpp
#define mrg_journal_jdir_hpp

namespace mrg
{
namespace journal
{
class jdir;
}
}

#include "jrnl/jinf.hpp"

namespace mrg
{
namespace journal
{

    /**
    * \class jdir
    * \brief Class to manage the journal directory
    */
    class jdir
    {
    private:
        std::string _dirname;
        std::string _base_filename;

    public:

        /**
        * \brief Sole constructor
        *
        * \param dirname Name of directory to be managed.
        * \param base_filename Filename root used in the creation of journal files
        *     and sub-directories.
        */
        jdir(const std::string& dirname, const std::string& base_filename);

        virtual ~jdir();


        /**
        * \brief Create journal directory as set in the dirname parameter of the constructor.
        *     Recursive creation is supported.
        *
        * \exception ??
        */
        void create_dir();

        /**
        * \brief Static function to create a directory. Recursive creation is supported.
        *
        * \param dirname C-string containing name of directory.
        *
        * \exception ??
        */
        static void create_dir(const char* dirname);

        /**
        * \brief Static function to create a directory. Recursive creation is supported.
        *
        * \param dirname String containing name of directory.
        *
        * \exception ??
        */
        static void create_dir(const std::string& dirname);


        /**
        * \brief Clear the journal directory of files matching the base filename
        *     by moving them into a subdirectory. This fn uses the dirname and base_filename
        *     that were set on construction.
        *
        * \param create_flag If set, create dirname if it is non-existent, otherwise throw
        *     exception.
        *
        * \exception The journal directory could not be opened.
        * \exception The move of files from the journal directory to the created backup
        *     directory failed.
        * \exception The directory handle could not be closed.
        */
        void clear_dir(const bool create_flag = true);

        /**
        * \brief Clear the directory dirname of journal files matching base_filename
        *     by moving them into a subdirectory.
        *
        * \param dirname C-string containing name of journal directory.
        * \param base_filename C-string containing base filename of journal files to be matched
        *     for moving into subdirectory.
        * \param create_flag If set, create dirname if it is non-existent, otherwise throw
        *     exception
        *
        * \exception The journal directory could not be opened.
        * \exception The move of files from the journal directory to the created backup
        *     directory failed.
        * \exception The directory handle could not be closed.
        */
        static void clear_dir(const char* dirname, const char* base_filename,
                const bool create_flag = true);

        /**
        * \brief Clear the directory dirname of journal files matching base_filename
        *     by moving them into a subdirectory.
        *
        * \param dirname String containing name of journal directory.
        * \param base_filename String containing base filename of journal files to be matched
        *     for moving into subdirectory.
        * \param create_flag If set, create dirname if it is non-existent, otherwise throw
        *     exception
        *
        * \exception The journal directory could not be opened.
        * \exception The move of files from the journal directory to the created backup
        *     directory failed.
        * \exception The directory handle could not be closed.
        */
        static void clear_dir(const std::string& dirname, const std::string& base_filename,
                const bool create_flag = true);


        /**
        * \brief ??
        *
        * \exception ??
        */
        void verify_dir();

        /**
        * \brief ??
        *
        * \param dirname C-string containing name of journal directory.
        * \param base_filename C-string containing base filename of journal files to be matched
        *     for moving into subdirectory.
        *
        * \exception ??
        */
        static void verify_dir(const char* dirname, const char* base_filename);

        /**
        * \brief ??
        *
        * \param dirname String containing name of journal directory.
        * \param base_filename String containing base filename of journal files to be matched
        *     for moving into subdirectory.
        *
        * \exception ??
        */
        static void verify_dir(const std::string& dirname, const std::string& base_filename);

        /**
        * \brief Delete the journal directory and all files and sub--directories that it may
        *     contain. This is equivilent of rm -rf.
        *
        * FIXME: links are not handled correctly.
        *
        * \param children_only To only delete children.
        *
        * \exception The journal directory could not be opened.
        * \exception The move of files from the journal directory to the created backup
        *     directory failed.
        * \exception The directory handle could not be closed.
        */
        void delete_dir(bool children_only = false );

        /**
        * \brief Delete the journal directory and all files and sub--directories that it may
        *     contain. This is equivilent of rm -rf.
        *
        * FIXME: links are not handled correctly.
        *
        * \param dirname C-string containing name of directory to be deleted.
        * \param children_only To only delete children.
        *
        * \exception The journal directory could not be opened.
        * \exception The move of files from the journal directory to the created backup
        *     directory failed.
        * \exception The directory handle could not be closed.
        */
        static void delete_dir(const char* dirname, bool children_only = false);

        /**
        * \brief Delete the journal directory and all files and sub--directories that it may
        *     contain. This is equivilent of rm -rf.
        *
        * FIXME: links are not handled correctly.
        *
        * \param dirname String containing name of directory to be deleted.
        * \param children_only To only delete children.
        *
        * \exception The journal directory could not be opened.
        * \exception The move of files from the journal directory to the created backup
        *     directory failed.
        * \exception The directory handle could not be closed.
        */
        static void delete_dir(const std::string& dirname, bool children_only = false);

        /**
        * \brief Create bakup directory that is next in sequence and move all journal files
        *     matching base_filename into it.
        *
        * In directory dirname, search for existing backup directory using pattern
        * "_basename.bak.XXXX" where XXXX is a hexadecimal sequence, and create next directory
        * based on highest number found. Move all journal files which match the base_fileaname
        * parameter into this new backup directory.
        *
        * \param dirname String containing name of journal directory.
        * \param base_filename String containing base filename of journal files to be matched
        *     for moving into subdirectory.
        *
        * \exception ??
        */
        static std::string create_bak_dir(const std::string& dirname,
                const std::string& base_filename);

        /**
        * \brief Return the directory name as a string.
        */
        inline const std::string& dirname() const { return _dirname; }

        /**
        * \brief Return the journal base filename name as a string.
        */
        inline const std::string& base_filename() const { return _base_filename; }

        /**
        * \brief Test whether the named file is a directory.
        *
        * \param name Name of file to be tested.
        * \return <b><i>true</i></b> if the named file is a directory; <b><i>false</i></b>
        *     otherwise.
        * \exception The stat() operation failed on the named file.
        */
        static bool is_dir(const char* name);

        /**
        * \brief Test whether the named file is a directory.
        *
        * \param name Name of file to be tested.
        * \return <b><i>true</i></b> if the named file is a directory; <b><i>false</i></b>
        *     otherwise.
        * \exception The stat() operation failed on the named file.
        */
        static bool is_dir(const std::string& name);


        /**
        * \brief Test whether the named entity exists on the filesystem.
        *
        * If stat() fails with error ENOENT, then this will return <b><i>false</i></b>. If
        * stat() succeeds, then <b><i>true</i></b> is returned, irrespective of the file type.
        * If stat() fails with any other error, an exception is thrown.
        *
        * \param name Name of entity to be tested.
        * \return <b><i>true</i></b> if the named entity exists; <b><i>false</i></b>
        *     otherwise.
        * \exception The stat() operation failed on the named entity.
        */
        static bool exists(const char* name);

        /**
        * \brief Test whether the named entity exists on the filesystem.
        *
        * If stat() fails with error ENOENT, then this will return <b><i>false</i></b>. If
        * stat() succeeds, then <b><i>true</i></b> is returned, irrespective of the file type.
        * If stat() fails with any other error, an exception is thrown.
        *
        * \param name Name of entity to be tested.
        * \return <b><i>true</i></b> if the named entity exists; <b><i>false</i></b>
        *     otherwise.
        * \exception The stat() operation failed on the named entity.
        */
        static bool exists(const std::string& name);

    public:

        /**
        * \brief Stream operator
        */
        friend std::ostream& operator<<(std::ostream& os, const jdir& jdir);

        /**
        * \brief Stream operator
        */
        friend std::ostream& operator<<(std::ostream& os, const jdir* jdirPtr);
    };

} // namespace journal
} // namespace mrg

#endif // ifndef mrg_journal_jdir_hpp
